library wireworld;

import 'dart:html';
import 'dart:async';

CanvasElement canvas;
CanvasRenderingContext2D context;
List<List<int>> grid, gridNew;
Timer running;
Point currentCellPosition, previousCellPosition;

int width = 480, height = 480, size = 30;
int TPS = 2; // ticks per second

void main() {
  
  previousCellPosition = new Point(0, 0);
  
  canvas = new CanvasElement(width: width, height: height);
  querySelector('#world').children.add(canvas);
  
  context = canvas.getContext("2d");
  canvas  
    ..onMouseMove.listen((event) => onMouseMove(event))
    ..onMouseDown.listen((event) => onMouseDown(event))
    ..onMouseUp.listen((event) => onMouseUp(event))
    ..onContextMenu.listen((event) => event.preventDefault());
   
  querySelector('#run').onClick.listen((event) => toggleRunning());
  querySelector('#clear').onClick.listen((event) => clear());
  querySelector('#cleare').onClick.listen((event) => clearElectrons());
  querySelector('#mirrorv').onClick.listen((event) => mirrorVertically());
  querySelector('#mirrorh').onClick.listen((event) => mirrorHorizontally());
  querySelector('#rotatec').onClick.listen((event) => rotateClockwise());
  querySelector('#rotatecc').onClick.listen((event) => rotateCounterclockwise());
  
  // setup grid
  grid = new List<List<int>>(size);
  gridNew = new List<List<int>>(size);
  for (var x = 0; x < size; x++) {
    grid[x] = new List<int>(size);
    gridNew[x] = new List<int>(size);
    for (var y = 0; y < size; y++) {
      grid[x][y] = 0;
      gridNew[x][y] = 0;
    }
  }
  
  window.requestAnimationFrame(draw);
}

void copyGrid() {
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) {     
      grid[x][y] = gridNew[x][y];
    }
  }  
}

void toggleRunning() {
  var value = querySelector("#run").attributes['value'];  
  if (value == "Run") {
    querySelector("#run").attributes['value'] = "Stop";
    running = new Timer.periodic(new Duration(milliseconds: 1000 ~/ TPS), (Timer timer) => update());
  } else {
    querySelector("#run").attributes['value'] = "Run";
    running.cancel();
  }
}

void clear() {
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) {    
      grid[x][y] = 0;
    }
  }
}

void clearElectrons() {
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) { 
      if (grid[x][y] == 2 || grid[x][y] == 3)
      grid[x][y] = 1;
    }
  }
}

void mirrorVertically() {
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) {     
      gridNew[x][y] = grid[size - 1 - x][y];
    }
  }  
  
  copyGrid(); 
}

void mirrorHorizontally() {
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) {     
      gridNew[x][y] = grid[x][size - 1 - y];
    }
  }  
  copyGrid();
}

void rotateClockwise() {
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) {  
      gridNew[size - 1 - y][x] = grid[x][y];
    }
  }
  
  copyGrid();
}

void rotateCounterclockwise() {
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) {    
      gridNew[y][size - 1 - x] = grid[x][y];
    }
  }
  
  copyGrid();
}

void update() {
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) {  
      gridNew[x][y] = grid[x][y];
    }
  }  
  
  // update cells according to wireworld rules
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) {      
      if (grid[x][y] == 1) { // copper to electron head
        int neighbours = 0;
        
        for (var i = -1; i <= 1; i++) {
          for (var j = -1; j <= 1; j++) {
            
           if (i != 0 || j != 0) {
             Point pos = new Point(x + i, y + j);
             if (pos.x > -1 && pos.x < size && pos.y > -1 && pos.y < size) {
               if (grid[pos.x][pos.y] == 2)
                 neighbours++;
             }
           }
          }
        }
        
        if (neighbours > 0 && neighbours <= 2)
          gridNew[x][y] = 2;
      }
      else if (grid[x][y] == 2) { // electron head to tail
        gridNew[x][y] = 3;
      }
      else if (grid[x][y] == 3) { // electron tail to copper
        gridNew[x][y] = 1;
      }
    }
  }
  
  copyGrid();
}

void draw(num _) {
  
  for (var x = 0; x < size; x++) {
    for (var y = 0; y < size; y++) { 
      
      if (grid[x][y] == 0)
        context.fillStyle = '#666'; // empty
      else if (grid[x][y] == 1)
        context.fillStyle = '#FF8000'; // copper
      else if (grid[x][y] == 2)
        context.fillStyle = '#fff'; // electron head
      else if (grid[x][y] == 3)
        context.fillStyle = '#0080FF'; // electron tail
      
      context
        ..beginPath()
        ..rect(x * 16, y * 16, 16, 16)
        ..fill()
        ..lineWidth = 1
        ..strokeStyle = '#333'
        ..stroke();
    }
  }
  
  // draw hover box
  if (currentCellPosition != null) {
    context
      ..beginPath()
      ..rect(currentCellPosition.x * 16, currentCellPosition.y * 16, 16, 16)
      ..lineWidth = 1
      ..strokeStyle = '#99F'
      ..stroke();
  }
  
  window.requestAnimationFrame(draw);
}

void updateMouse(MouseEvent evt) {
  Point mousePosition = new Point(evt.client.x - canvas.getBoundingClientRect().left,
                                  evt.client.y - canvas.getBoundingClientRect().top);
  
  currentCellPosition = new Point(mousePosition.x ~/ 16, mousePosition.y ~/ 16);
}

void onMouseMove(MouseEvent evt) {
  updateMouse(evt);
  setCell(evt);
    
  if (evt.which != 0)
    if (currentCellPosition.x != previousCellPosition.x || currentCellPosition.y != previousCellPosition.y) {  
      previousCellPosition = currentCellPosition;
    }
}

void onMouseDown(MouseEvent evt) {  
  evt.preventDefault(); // prevent the mouse cursor from changing
}

void onMouseUp(MouseEvent evt) {  
  setCell(evt);
}

void setCell(MouseEvent evt) {      
  if (evt.which == 1) {
    if (currentCellPosition.x != previousCellPosition.x || currentCellPosition.y != previousCellPosition.y) {
      if (grid[currentCellPosition.x][currentCellPosition.y] != 1)
        grid[currentCellPosition.x][currentCellPosition.y] = 1; // copper
      else
        grid[currentCellPosition.x][currentCellPosition.y] = 0; // empty
    }
  }
  else if (evt.which == 2) { // electron head
    grid[currentCellPosition.x][currentCellPosition.y] = 2; 
  }
  else if (evt.which == 3) { // electron tail
    grid[currentCellPosition.x][currentCellPosition.y] = 3;
  } 
}